import json
from urllib.parse import urljoin

import requests

from cancelchain.util import dt_2_ciso, host_address

OK = requests.codes.ok
UNAUTHORIZED = requests.codes.unauthorized
PEER_HOST_HEADER = 'Peer-Hosts'
ADDRESS_MISMATCH_MSG = 'Address/wallet mismatch'


def json_header(headers=None):
    headers = headers or {}
    headers['Content-Type'] = 'application/json'
    return headers


def peer_header(visited_hosts, headers=None):
    headers = headers or {}
    if visited_hosts:
        headers[PEER_HOST_HEADER] = ','.join(visited_hosts)
    return headers


class ApiClient():
    def __init__(self, host, wallet, timeout=None):
        host, address = host_address(host)
        if address and address != wallet.address:
            raise Exception(ADDRESS_MISMATCH_MSG)
        self.host = host
        self.wallet = wallet
        self.token = None
        self.timeout = timeout if timeout is not None else 10

    def request_token(self, rfs=True):
        r = requests.get(
            urljoin(self.host, f'/api/token/{self.wallet.address}'),
            timeout=self.timeout
        )
        if rfs:
            r.raise_for_status()
        if r.status_code == OK:
            secret = self.wallet.decrypt(r.json().get('cipher')).decode()
            r = requests.post(
                urljoin(self.host, f'/api/token/{self.wallet.address}'),
                headers=json_header(),
                data=json.dumps({'challenge': secret}),
                timeout=self.timeout
            )
            if rfs:
                r.raise_for_status()
            if r.status_code == OK:
                return r.json().get('token')
        return None

    def get_token(self, rfs=True):
        if self.token is None:
            self.token = self.request_token(rfs=rfs)
        return self.token

    def reset_token(self):
        self.token = None

    def auth_header(self, headers=None, rfs=True):
        headers = headers or {}
        token = self.get_token(rfs=rfs)
        if token:
            headers['Authorization'] = f'Bearer {token}'
        return headers

    def get(
        self, path,
        headers=None,
        params=None,
        timeout=None,
        raise_for_status=True
    ):
        timeout = self.timeout if timeout is None else timeout
        for _i in range(2):
            headers = self.auth_header(headers=headers, rfs=raise_for_status)
            r = requests.get(
                urljoin(self.host, path),
                headers=headers,
                params=params,
                timeout=timeout
            )
            if r.status_code == UNAUTHORIZED:
                self.reset_token()
            else:
                break
        if raise_for_status:
            r.raise_for_status()
        return r

    def post(
        self, path,
        headers=None,
        data=None,
        timeout=None,
        raise_for_status=True
    ):
        timeout = self.timeout if timeout is None else timeout
        for _i in range(2):
            headers = self.auth_header(headers=headers, rfs=raise_for_status)
            r = requests.post(
                urljoin(self.host, path),
                headers=headers,
                data=data,
                timeout=timeout
            )
            if r.status_code == UNAUTHORIZED:
                self.reset_token()
            else:
                break
        if raise_for_status:
            r.raise_for_status()
        return r

    def get_transfer_transaction(
        self, public_key, amount, address, timeout=None, raise_for_status=True
    ):
        return self.get(
            '/api/transaction/transfer',
            params={
                'public_key': public_key,
                'amount': amount,
                'address': address
            },
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_subject_transaction(
        self, public_key, amount, subject, timeout=None, raise_for_status=True
    ):
        return self.get(
            '/api/transaction/subject',
            params={
                'public_key': public_key,
                'amount': amount,
                'subject': subject
            },
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_forgive_transaction(
        self, public_key, amount, subject, timeout=None, raise_for_status=True
    ):
        return self.get(
            '/api/transaction/forgive',
            params={
                'public_key': public_key,
                'amount': amount,
                'subject': subject
            },
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_support_transaction(
        self, public_key, amount, subject, timeout=None, raise_for_status=True
    ):
        return self.get(
            '/api/transaction/support',
            params={
                'public_key': public_key,
                'amount': amount,
                'subject': subject
            },
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def post_transaction(
        self, txn, visited_hosts=None, timeout=None, raise_for_status=True
    ):
        headers = peer_header(visited_hosts, headers=json_header())
        return self.post(
            f'/api/transaction/{txn.txid}',
            data=txn.to_json(),
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_pending_transactions(
        self, earliest=None, timeout=None, raise_for_status=True
    ):
        params = None
        if earliest is not None:
            params = {'earliest': dt_2_ciso(earliest)}
        return self.get(
            '/api/transaction/pending',
            params=params,
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_block(self, block_hash=None, timeout=None, raise_for_status=True):
        return self.get(
            f'/api/block/{block_hash}' if block_hash else '/api/block',
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def post_block(
        self, block, visited_hosts=None, timeout=None, raise_for_status=True
    ):
        headers = peer_header(visited_hosts, headers=json_header())
        return self.post(
            f'/api/block/{block.block_hash}',
            data=block.to_json(),
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_wallet_balance(self, address, timeout=None, raise_for_status=True):
        return self.get(
            f'/api/wallet/{address}/balance',
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_subject_balance(
        self, subject, timeout=None, raise_for_status=True
    ):
        return self.get(
            f'/api/subject/{subject}/balance',
            timeout=timeout,
            raise_for_status=raise_for_status
        )

    def get_subject_support(
        self, subject, timeout=None, raise_for_status=True
    ):
        return self.get(
            f'/api/subject/{subject}/support',
            timeout=timeout,
            raise_for_status=raise_for_status
        )



from base64 import urlsafe_b64decode, urlsafe_b64encode
from dataclasses import dataclass

from marshmallow import (
    ValidationError,
    fields,
    post_load,
    validate,
    validates_schema,
)

from cancelchain.schema import Address, MillHash, SansNoneSchema

MIN_SUBJECT_LENGTH = 1
MAX_SUBJECT_LENGTH = 79
INVALID_DESTINATION_MSG = 'Invalid destinations'
INVALID_PADDING_MSG = 'Invalid padding'


def encode_subject(raw_subject):
    return urlsafe_b64encode(raw_subject.encode()).rstrip(b'=').decode()


def decode_subject(subject):
    if subject.endswith('='):
        raise TypeError(INVALID_PADDING_MSG)
    subject = subject.encode()
    subject += b'=' * (-len(subject) % 4)
    return urlsafe_b64decode(subject).decode()


def validate_subject(subject):
    try:
        raw_subject = decode_subject(subject)
        if MIN_SUBJECT_LENGTH <= len(raw_subject) <= MAX_SUBJECT_LENGTH:
            return encode_subject(raw_subject) == subject
    except Exception:
        pass
    return False


def validate_raw_subject(raw_subject):
    try:
        if MIN_SUBJECT_LENGTH <= len(raw_subject) <= MAX_SUBJECT_LENGTH:
            return decode_subject(encode_subject(raw_subject)) == raw_subject
    except Exception:
        pass
    return False


class Subject(fields.String):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.validators.insert(0, validate_subject)


class OutflowSchema(SansNoneSchema):
    amount = fields.Integer(required=True, validate=validate.Range(min=1))
    address = Address()
    subject = Subject()
    forgive = Subject()
    support = Subject()

    @validates_schema
    def validate_destinations(self, data, **kwargs):
        address = data.get('address')
        options = [
            v for v in [
                data.get(n) for n in ('subject', 'forgive', 'support')
            ] if v is not None
        ]
        if not (
            (address and not options) or
            (options and len(options) == 1 and not address)
        ):
            raise ValidationError(INVALID_DESTINATION_MSG)

    @post_load
    def make_outflow(self, data, **kwargs):
        return Outflow(**data)


@dataclass
class Outflow():
    amount: int = None
    address: str = None
    subject: str = None
    forgive: str = None
    support: str = None

    @property
    def data_csv(self):
        return ','.join([
            str(self.amount),
            self.address if self.address is not None else '',
            self.subject if self.subject is not None else '',
            self.forgive if self.forgive is not None else '',
            self.support if self.support is not None else ''
        ])

    @property
    def schadenfreude(self):
        return int(self.amount / 2) if self.subject is not None else 0

    @property
    def grace(self):
        return int(self.amount / 2) if self.forgive is not None else 0

    @property
    def mudita(self):
        return self.amount if self.support is not None else 0


class InflowSchema(SansNoneSchema):
    outflow_txid = MillHash(required=True)
    outflow_idx = fields.Integer(required=True, validate=validate.Range(min=0))

    @post_load
    def make_inflow(self, data, **kwargs):
        return Inflow(**data)


@dataclass
class Inflow:
    outflow_txid: str = None
    outflow_idx: int = None

    @property
    def data_csv(self):
        return ','.join([self.outflow_txid, str(self.outflow_idx)])



import json
import re
from enum import Enum
from functools import wraps
from urllib.parse import urljoin

import jwt
from flask import Blueprint, abort, current_app, make_response, request
from flask.views import MethodView
from marshmallow import Schema, ValidationError, fields, validate

from cancelchain.api_client import PEER_HOST_HEADER, ApiClient
from cancelchain.block import TXN_TIMEOUT, Block
from cancelchain.cache import cache
from cancelchain.exceptions import CCError, EmptyChainError, MissingBlockError
from cancelchain.models import ApiToken
from cancelchain.node import Node
from cancelchain.payload import encode_subject, validate_raw_subject
from cancelchain.schema import validate_address_format, validate_public_key
from cancelchain.signals import http_post as http_post_signal
from cancelchain.tasks import post_process
from cancelchain.util import ciso_2_dt, dt_2_ciso, host_address, now, now_iso
from cancelchain.wallet import Wallet

API_TOKEN_SECONDS = 60 * 60 * 4

blueprint = Blueprint('api', __name__)


def node_lc_dao():
    node = Node(
        host=current_app.config['NODE_HOST'],
        peers=current_app.config['PEERS'],
        clients=current_app.clients,
        logger=current_app.logger
    )
    lc = node.longest_chain
    return node, lc, lc.to_dao() if lc is not None else None


def visited_hosts():
    hosts = None
    if peer_hosts := request.headers.get(PEER_HOST_HEADER, None):
        hosts = [v.strip() for v in peer_hosts.split(',') if v]
    return hosts


def queue_post_process(path, data, visited_hosts):
    host, address = host_address(current_app.config['NODE_HOST'])
    wallet = current_app.wallets.get(address)
    headers = None
    if visited_hosts:
        headers = {PEER_HOST_HEADER: ','.join(visited_hosts)}
    headers = ApiClient(host, wallet).auth_header(headers=headers)
    url = urljoin(host, path)
    http_post_signal.send(
        current_app._get_current_object(), url=url, data=data, headers=headers
    )


def queue_block_post_process(block, visited_hosts):
    queue_post_process(
        f'/api/block/{block.block_hash}/process',
        block.to_json(),
        visited_hosts
    )


def queue_txn_post_process(txn, visited_hosts):
    queue_post_process(
        f'/api/transaction/{txn.txid}/process',
        txn.to_json(),
        visited_hosts
    )


def handle_http_post(sender, url=None, data=None, headers=None):
    if current_app.config.get('CELERY_BROKER_URL'):
        post_process.delay(url, data, headers=headers)


@blueprint.record
def connect_signals(state):
    http_post_signal.connect(handle_http_post)


def make_json_response(json_data, status_code=200):
    if not isinstance(json_data, (str, bytes)):
        json_data = json.dumps(json_data)
    response = make_response(json_data, status_code)
    response.headers['Content-Type'] = 'application/json'
    return response


def make_error_response(e):
    return make_json_response({'error': e.messages}, 400)


def exception_response(e):
    current_app.logger.exception(e)
    abort(500)


class Role(Enum):
    READER = 1
    TRANSACTOR = 2
    MILLER = 3
    ADMIN = 4

    def addresses(self):
        return current_app.config.get(f'{self.name}_ADDRESSES')

    @classmethod
    def address_roles(cls, address):
        return [role for role in Role if any(
            re.fullmatch(x, address) for x in role.addresses()
        )]

    @classmethod
    def address_role(cls, address):
        roles = cls.address_roles(address)
        return roles[-1] if roles else None


class TokenView(MethodView):
    def get(self, address):
        api_token = ApiToken.get(address)
        if not api_token:
            if not (wallet := current_app.wallets.get(address)):
                _, _, lc_dao = node_lc_dao()
                if lc_dao is None:
                    abort(401)
                if txn := lc_dao.address_transactions(address).first():
                    wallet = Wallet(b64ks=txn.public_key)
            if not wallet:
                abort(401)
            api_token = ApiToken.create(wallet)
        return make_json_response({'cipher': api_token.refreshed_cipher()})

    def post(self, address):
        if (api_token := ApiToken.get(address)) is None:
            abort(401)
        if not api_token.verify(request.json.get('challenge')):
            abort(401)
        api_token.reset()
        if (role := Role.address_role(address)) is None:
            abort(403)
        token = jwt.encode(
            {
                'sub': address,
                'rol': str(role.name),
                'exp': now().timestamp() + API_TOKEN_SECONDS
            },
            current_app.config['SECRET_KEY'],
            algorithm='HS256'
        )
        return make_json_response({'token': token})


blueprint.add_url_rule(
    '/token/<address:address>',
    view_func=TokenView.as_view('token'),
    methods=['GET', 'POST']
)


def authorize(required_role=Role.READER):
    def _authorize(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            authorized = False
            address = None
            role = None
            try:
                token = request.headers.get('Authorization')
                if token and token.startswith('Bearer '):
                    token = token[7:]
                else:
                    token = None
                if token:
                    data = jwt.decode(
                        token,
                        current_app.config['SECRET_KEY'],
                        algorithms=['HS256']
                    )
                    address = data['sub']
                    role = Role[data['rol']]
                    if address and role.value >= required_role.value:
                        authorized = True
            except jwt.exceptions.ExpiredSignatureError:
                abort(401)
            except Exception as e:
                current_app.logger.exception(e)
                abort(401)
            if authorized:
                kwargs['_address'] = address
                kwargs['_role'] = role
                return func(*args, **kwargs)
            abort(401)
        return wrapper
    return _authorize


authorize_reader = authorize(required_role=Role.READER)
authorize_transactor = authorize(required_role=Role.TRANSACTOR)
authorize_miller = authorize(required_role=Role.MILLER)
authorize_admin = authorize(required_role=Role.ADMIN)


class BlockView(MethodView):
    def get(self, block_hash=None, **kwargs):
        try:
            _, lc, _ = node_lc_dao()
            block = None
            if not block_hash and lc is not None:
                block = lc.last_block if lc else None
                block_hash = block.block_hash if block else None
            if block_hash:
                key = f'{block_hash}.block-json'
                if (block_json := cache.get(key)) is None:
                    block = block or Block.from_db(block_hash)
                    if block is not None:
                        block_json = block.to_json()
                        cache.set(key, block_json)
                if block_json:
                    return make_json_response(block_json)
        except CCError as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)
        abort(404)

    def post(self, block_hash, process=False, **kwargs):
        try:
            process = process == 'process'
            if not process:
                process = not current_app.config.get('API_ASYNC_PROCESSING')
            node, _, _ = node_lc_dao()
            vhosts = visited_hosts()
            received = now_iso()
            block = node.receive_block(
                request.data, block_hash=block_hash,
                visited_hosts=vhosts, process=process
            )
            if process is False and block is not None:
                queue_block_post_process(block, vhosts)
        except MissingBlockError:
            abort(404)
        except CCError as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)
        status_code = 200 if block is None else 201 if process else 202
        return make_json_response(
            {'received': received}, status_code=status_code
        )


reader_block_view = authorize_reader(BlockView.as_view('block_reader'))
miller_block_view = authorize_miller(BlockView.as_view('block_miller'))

blueprint.add_url_rule(
    '/block',
    view_func=reader_block_view,
    methods=['GET']
)

blueprint.add_url_rule(
    '/block/<mill_hash:block_hash>',
    view_func=reader_block_view,
    methods=['GET']
)

blueprint.add_url_rule(
    '/block/<mill_hash:block_hash>',
    view_func=miller_block_view,
    methods=['POST']
)

blueprint.add_url_rule(
    '/block/<mill_hash:block_hash>/<process>',
    view_func=miller_block_view,
    methods=['POST']
)


class TxnView(MethodView):
    def post(self, txid, process=False, **kwargs):
        try:
            process = process == 'process'
            if not process:
                process = not current_app.config.get('API_ASYNC_PROCESSING')
            node, _, _ = node_lc_dao()
            vhosts = visited_hosts()
            received = now_iso()
            txn = node.receive_transaction(
                txid, request.data, visited_hosts=vhosts, process=process
            )
            if process is False and txn is not None:
                queue_txn_post_process(txn, vhosts)
        except CCError as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)
        status_code = 200 if txn is None else 201 if process else 202
        return make_json_response(
            {'received': received}, status_code=status_code
        )


transactor_txn_view = authorize_transactor(TxnView.as_view('txn_transactor'))

blueprint.add_url_rule(
    '/transaction/<mill_hash:txid>',
    view_func=transactor_txn_view,
    methods=['POST']
)
blueprint.add_url_rule(
    '/transaction/<mill_hash:txid>/<process>',
    view_func=transactor_txn_view,
    methods=['POST']
)


class TransferTxnQuerySchema(Schema):
    public_key = fields.String(required=True, validate=validate_public_key)
    amount = fields.Integer(required=True, validate=validate.Range(min=1))
    address = fields.String(required=True, validate=validate_address_format)


class TransferTxnView(MethodView):
    def get(self, **kwargs):
        try:
            args = TransferTxnQuerySchema().load(request.args)
            public_key_b64 = args['public_key']
            amount = args['amount']
            dest_address = args['address']
            wallet = Wallet(b64ks=public_key_b64)
            _, lc, _ = node_lc_dao()
            if lc is None:
                raise EmptyChainError()
            return make_json_response(
                lc.create_transfer(wallet, amount, dest_address).to_json()
            )
        except (ValidationError, CCError) as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/transaction/transfer',
    view_func=authorize_transactor(
        TransferTxnView.as_view('txn_transfer_transactor')
    ),
    methods=['GET']
)


class SubjectTxnQuerySchema(Schema):
    public_key = fields.String(required=True, validate=validate_public_key)
    amount = fields.Integer(required=True, validate=validate.Range(min=1))
    subject = fields.String(required=True, validate=validate_raw_subject)


class SubjectTxnView(MethodView):
    def get(self, **kwargs):
        try:
            args = SubjectTxnQuerySchema().load(request.args)
            public_key_b64 = args['public_key']
            amount = args['amount']
            subject = encode_subject(args['subject'])
            wallet = Wallet(b64ks=public_key_b64)
            _, lc, _ = node_lc_dao()
            if lc is None:
                raise EmptyChainError()
            return make_json_response(
                lc.create_subject(wallet, amount, subject).to_json()
            )
        except (ValidationError, CCError) as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/transaction/subject',
    view_func=authorize_transactor(
        SubjectTxnView.as_view('txn_subject_transactor')
    ),
    methods=['GET']
)


class ForgiveTxnView(MethodView):
    def get(self, **kwargs):
        try:
            args = SubjectTxnQuerySchema().load(request.args)
            public_key_b64 = args['public_key']
            amount = args['amount']
            subject = encode_subject(args['subject'])
            wallet = Wallet(b64ks=public_key_b64)
            _, lc, _ = node_lc_dao()
            if lc is None:
                raise EmptyChainError()
            return make_json_response(
                lc.create_forgive(wallet, amount, subject).to_json()
            )
        except (ValidationError, CCError) as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/transaction/forgive',
    view_func=authorize_transactor(
        ForgiveTxnView.as_view('txn_forgive_transactor')
    ),
    methods=['GET']
)


class SupportTxnView(MethodView):
    def get(self, **kwargs):
        try:
            args = SubjectTxnQuerySchema().load(request.args)
            public_key_b64 = args['public_key']
            amount = args['amount']
            subject = encode_subject(args['subject'])
            wallet = Wallet(b64ks=public_key_b64)
            _, lc, _ = node_lc_dao()
            if lc is None:
                raise EmptyChainError()
            return make_json_response(
                lc.create_support(wallet, amount, subject).to_json()
            )
        except (ValidationError, CCError) as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/transaction/support',
    view_func=authorize_transactor(
        SupportTxnView.as_view('txn_support_transactor')
    ),
    methods=['GET']
)


class PendingTxnQuerySchema(Schema):
    earliest = fields.Function(
        lambda obj: dt_2_ciso(obj.earliest),
        deserialize=lambda v: ciso_2_dt(v),
        required=False
    )


class PendingTxnView(MethodView):
    def get(self, **kwargs):
        try:
            node, _, _ = node_lc_dao()
            node.discard_expired_pending_txns()
            args = PendingTxnQuerySchema().load(request.args)
            earliest = args.get('earliest')
            expired = now() - TXN_TIMEOUT
            pending_json = node.pending_txns.query_json(
                earliest=earliest, expired=expired
            )
            return make_json_response([json.loads(j) for j in pending_json])
        except (ValidationError, CCError) as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/transaction/pending',
    view_func=authorize_reader(
        PendingTxnView.as_view('txn_pending_reader')
    ),
    methods=['GET']
)


class WalletBalanceView(MethodView):
    def get(self, address, **kwargs):
        try:
            _, lc, _ = node_lc_dao()
            if lc is None:
                raise EmptyChainError()
            block_hash = lc.block_hash
            key = f'{block_hash}.{address}.wallet-balance'
            if (balance := cache.get(key)) is None:
                balance = lc.balance(address)
                cache.set(key, balance)
            return make_json_response(
                {'balance': balance, 'as_of_block': block_hash}
            )
        except CCError as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/wallet/<address:address>/balance',
    view_func=authorize_reader(
        WalletBalanceView.as_view('wallet_balance_transactor')
    ),
    methods=['GET']
)


class SubjectBalanceView(MethodView):
    def get(self, subject, **kwargs):
        try:
            _, lc, _ = node_lc_dao()
            if lc is None:
                raise EmptyChainError()
            block_hash = lc.block_hash
            key = f'{block_hash}.{subject}.balance'
            if (balance := cache.get(key)) is None:
                balance = lc.subject_balance(subject)
                cache.set(key, balance)
            return make_json_response(
                {'balance': balance, 'as_of_block': block_hash}
            )
        except CCError as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/subject/<subject:subject>/balance',
    view_func=authorize_reader(
        SubjectBalanceView.as_view('subject_balance_transactor')
    ),
    methods=['GET']
)


class SubjectSupportView(MethodView):
    def get(self, subject, **kwargs):
        try:
            _, lc, _ = node_lc_dao()
            if lc is None:
                raise EmptyChainError()
            block_hash = lc.block_hash
            key = f'{block_hash}.{subject}.support'
            if (support := cache.get(key)) is None:
                support = lc.subject_support(subject)
                cache.set(key, support)
            return make_json_response(
                {'support': support, 'as_of_block': block_hash}
            )
        except CCError as err:
            return make_error_response(err)
        except Exception as e:
            exception_response(e)


blueprint.add_url_rule(
    '/subject/<subject:subject>/support',
    view_func=authorize_reader(
        SubjectSupportView.as_view('subject_support_transactor')
    ),
    methods=['GET']
)



from dataclasses import dataclass, field
from datetime import timedelta
from json import JSONDecodeError

from marshmallow import (
    ValidationError,
    fields,
    post_load,
    validate,
    validates_schema,
)
from pymerkle import MerkleTree, verify_inclusion
from pymerkle.proof import InvalidProof

from cancelchain.exceptions import (
    ExpiredTransactionError,
    FutureTransactionError,
    InvalidBlockError,
    InvalidBlockHashError,
    InvalidCoinbaseError,
    InvalidMerkleRootError,
    InvalidProofError,
    InvalidTransactionError,
    MissingCoinbaseError,
    OutOfOrderTransactionError,
    SealedBlockError,
    UnlinkedBlockError,
)
from cancelchain.milling import mill_hash_str, milling_generator
from cancelchain.models import BlockDAO
from cancelchain.schema import (
    MillHash,
    SansNoneSchema,
    Timestamp,
    asdict_sans_none,
)
from cancelchain.transaction import Transaction, TransactionSchema
from cancelchain.util import dt_2_iso, iso_2_dt, now_iso

VERSION_1 = '1'
MAX_TRANSACTIONS = 100
TXN_TIMEOUT = timedelta(hours=4)
MISSED_TARGET_MSG = 'Missed target'


def validate_hash_diff(block_hash, target):
    return int(block_hash, 16) < int(target, 16)


class BlockSchema(SansNoneSchema):
    idx = fields.Integer(required=True, validate=validate.Range(min=0))
    timestamp = Timestamp(required=True)
    block_hash = MillHash(required=True)
    prev_hash = MillHash(required=True)
    target = MillHash(required=True)
    proof_of_work = fields.Integer(
        required=True, validate=validate.Range(min=0)
    )
    merkle_root = MillHash(required=True)
    txns = fields.List(
        fields.Nested(TransactionSchema),
        required=True,
        validate=validate.Length(min=1, max=MAX_TRANSACTIONS)
    )
    version = fields.String(required=True, validate=validate.Equal(VERSION_1))

    @validates_schema
    def validate_difficulty(self, data, **kwargs):
        if not validate_hash_diff(data.get('block_hash'), data.get('target')):
            raise ValidationError(MISSED_TARGET_MSG)

    @post_load
    def make_block(self, data, **kwargs):
        return Block(**data)


@dataclass(order=True)
class Block:
    idx: int = field(default=None)
    timestamp: str = field(default=None)
    block_hash: str = field(default=None)
    prev_hash: str = field(default=None, compare=False)
    target: str = field(default='F' * 64, compare=False, repr=False)
    proof_of_work: int = field(default=None, compare=False, repr=False)
    merkle_root: str = field(default=None, compare=False, repr=False)
    txns: list[Transaction] = field(default_factory=list, compare=False)
    version: str = field(default=VERSION_1, compare=False, repr=False)

    @property
    def timestamp_dt(self):
        return iso_2_dt(self.timestamp) if self.timestamp else None

    @property
    def last_txn(self):
        return self.txns[-1] if self.txns else None

    @property
    def regular_txns(self):
        return self.txns[0:-1] if self.txns else []

    @property
    def coinbase(self):
        return self.last_txn if self.is_sealed else None

    @property
    def schadenfreude(self):
        return sum([t.schadenfreude for t in self.txns])

    @property
    def grace(self):
        return sum([t.grace for t in self.txns])

    @property
    def mudita(self):
        return sum([t.mudita for t in self.txns])

    @property
    def is_sealed(self):
        return self.timestamp is not None

    @property
    def is_proved(self):
        if self.block_hash:
            return validate_hash_diff(self.block_hash, self.target)
        return False

    @property
    def unproven_header(self):
        return ','.join((
            str(self.idx),
            str(self.timestamp),
            str(self.prev_hash),
            str(self.target),
            str(self.merkle_root),
            str(self.version),
            ''
        ))

    @property
    def header(self):
        return self.potential_header(self.proof_of_work)

    def get_header_hash(self):
        return mill_hash_str(self.header)

    def potential_header(self, proof_of_work):
        return f'{self.unproven_header}{proof_of_work}'

    def validate_proof_of_work(self, proof_of_work):
        potential_header = self.potential_header(proof_of_work)
        return validate_hash_diff(mill_hash_str(potential_header), self.target)

    def build_merkle_tree(self):
        tree = MerkleTree()
        for record in (t.txid for t in self.txns):
            tree.append_entry(record)
        return tree

    def get_merkle_root(self):
        root_hash = self.build_merkle_tree().root
        return root_hash.decode() if root_hash else None

    def in_merkle_tree(self, txid):
        tree = self.build_merkle_tree()
        target = tree.root
        proof = tree.prove_inclusion(txid)
        try:
            verify_inclusion(txid, target, proof)
        except InvalidProof:
            return False
        else:
            return True

    def add_txn(self, txn, is_coinbase=False):
        if self.is_sealed:
            raise SealedBlockError()
        if not is_coinbase:
            self.validate_transaction(txn, prev_txn=self.last_txn)
        else:
            txn.validate_coinbase()
        self.txns.append(txn)

    def create_coinbase(self, wallet, reward):
        return Transaction.coinbase(
            wallet, reward, self.schadenfreude, self.grace, self.mudita
        )

    def add_coinbase(self, wallet, reward):
        self.add_txn(self.create_coinbase(wallet, reward), is_coinbase=True)

    def link(self, idx, prev_hash, target):
        self.idx = idx
        self.prev_hash = prev_hash
        self.target = target

    def seal(self, wallet, reward):
        if self.is_sealed:
            raise SealedBlockError()
        if (self.prev_hash is None) or (self.idx is None):
            raise UnlinkedBlockError()
        self.txns.sort()
        self.add_coinbase(wallet, reward)
        self.merkle_root = self.get_merkle_root()
        self.timestamp = now_iso()

    def mill(self, mp=False, progress=None):
        mg = milling_generator(self, mp=mp, progress=progress)
        while next(mg) is None:
            pass

    def solve(self, proof_of_work):
        if self.validate_proof_of_work(proof_of_work):
            self.proof_of_work = proof_of_work
            self.block_hash = self.get_header_hash()
        else:
            raise InvalidProofError()

    def validate_block_hash(self):
        if self.block_hash != self.get_header_hash():
            raise InvalidBlockHashError()

    def validate_merkle_root(self):
        if self.merkle_root != self.get_merkle_root():
            raise InvalidMerkleRootError()

    def validate_transaction(self, txn, prev_txn=None):
        txn.validate(coinbase=False)
        txn_ts_dt = txn.timestamp_dt
        if self.timestamp_dt and txn_ts_dt > self.timestamp_dt:
            raise FutureTransactionError()
        if self.timestamp_dt and txn_ts_dt < self.timestamp_dt - TXN_TIMEOUT:
            raise ExpiredTransactionError()
        if prev_txn and txn < prev_txn:
            raise OutOfOrderTransactionError()

    def validate_coinbase(self):
        cb = self.coinbase
        if not cb:
            raise MissingCoinbaseError()
        cb.validate_coinbase()
        comps = []
        if self.schadenfreude:
            comps.append(self.schadenfreude)
        if self.grace:
            comps.append(self.grace)
        if self.mudita:
            comps.append(self.mudita)
        if comps != [o.amount for o in cb.outflows[1:]]:
            raise InvalidCoinbaseError()

    def validate(self):
        if errors := BlockSchema().validate(self.to_dict()):
            raise InvalidBlockError(errors)
        self.validate_block_hash()
        self.validate_merkle_root()
        prev_txn = None
        for txn in self.regular_txns:
            try:
                self.validate_transaction(txn, prev_txn=prev_txn)
            except InvalidTransactionError as e:
                raise InvalidBlockError({f'Transaction {txn.txid}': e.messages})
            prev_txn = txn
        try:
            self.validate_coinbase()
        except InvalidTransactionError as e:
            raise InvalidBlockError(e.messages)

    def to_dict(self):
        return asdict_sans_none(self)

    def to_json(self):
        return BlockSchema().dumps(self.to_dict())

    def to_dao(self):
        return BlockDAO.get(self.block_hash) or BlockDAO(
            self.block_hash, self.version, self.idx, self.prev_hash,
            self.timestamp_dt, self.merkle_root, self.proof_of_work,
            self.target, transaction_daos=[txn.to_dao() for txn in self.txns]
        )

    def to_db(self):
        self.to_dao().commit()

    @classmethod
    def from_dict(cls, d):
        try:
            return BlockSchema().load(d)
        except ValidationError as e:
            raise InvalidBlockError(e.messages)

    @classmethod
    def from_json(cls, j):
        try:
            return BlockSchema().loads(j)
        except JSONDecodeError as je:
            raise InvalidBlockError(je.msg)
        except ValidationError as ve:
            raise InvalidBlockError(ve.messages)

    @classmethod
    def from_dao(cls, dao):
        return cls(
            idx=dao.idx,
            timestamp=dt_2_iso(dao.timestamp),
            block_hash=dao.block_hash,
            prev_hash=dao.prev_hash,
            target=dao.target,
            proof_of_work=dao.proof_of_work,
            merkle_root=dao.merkle_root,
            txns=[
                Transaction.from_dao(txn_dao) for txn_dao in dao.transactions
            ],
            version=dao.version
        )

    @classmethod
    def from_db(cls, block_hash):
        dao = BlockDAO.get(block_hash)
        return cls.from_dao(dao) if dao else None



import json
import os
from dataclasses import dataclass, field, fields
from typing import ClassVar


@dataclass
class EnvironSettings:
    _prefix: ClassVar[str] = ''

    @classmethod
    def getenv(cls, name):
        return os.environ.get(f'{cls._prefix}{name}')

    @classmethod
    def from_env(cls):
        c = cls()
        for f in fields(c):
            if (v := cls.getenv(f.name)) is not None:
                v = v.strip()
                try:
                    setattr(c, f.name, json.loads(v))
                except Exception:
                    setattr(c, f.name, v)
        return c


@dataclass
class EnvAppSettings(EnvironSettings):
    _prefix: ClassVar[str] = 'CC_'

    NODE_HOST: str = field(default=None)
    PEERS: list[str] = field(default_factory=list)
    API_CLIENT_TIMEOUT: int = field(default=10)
    API_ASYNC_PROCESSING: bool = field(default=False)
    DEFAULT_COMMAND_HOST: str = field(default=None)
    WALLET_DIR: str = field(default=None)
    ADMIN_ADDRESSES: list[str] = field(default_factory=list)
    MILLER_ADDRESSES: list[str] = field(default_factory=list)
    TRANSACTOR_ADDRESSES: list[str] = field(default_factory=list)
    READER_ADDRESSES: list[str] = field(default_factory=list)



import multiprocessing
from hashlib import sha256, sha512
from itertools import count


def mill_hash(data):
    if isinstance(data, str):
        data = data.encode()
    return sha256(sha512(data).digest())


def mill_hash_bin(data):
    return mill_hash(data).digest()


def mill_hash_str(data):
    return mill_hash(data).hexdigest()


def mill_work(w):
    work_start, work_stop, unproven_header, target = w
    for proof in range(work_start, work_stop):
        h = mill_hash_str(f'{unproven_header}{proof}')
        if int(h, 16) < target:
            return (proof, work_stop - proof)
    return (None, work_stop - work_start)


def mill_block(block, rounds, worksize, progress_next):
    target = int(block.target, 16)
    unproven_header = block.unproven_header
    proof_of_work = None
    proof_start = 0
    r = range(rounds) if rounds else count()
    while proof_of_work is None:
        for _i in r:
            if proof_of_work is not None:
                break
            proof, c = mill_work((
                proof_start, proof_start + worksize, unproven_header, target
            ))
            progress_next(n=c)
            if proof is not None and proof_of_work is None:
                proof_of_work = proof
            proof_start += worksize
        yield proof_of_work


def work_generator(unproven_header, target, start, worksize, num):
    for i in range(num):
        work_start = start + (i * worksize)
        yield (work_start, work_start + worksize, unproven_header, target)


def mill_block_mp(block, rounds, worksize, progress_next):
    cpus = multiprocessing.cpu_count()
    target = int(block.target, 16)
    unproven_header = block.unproven_header
    proof_of_work = None
    proof_start = 0
    r = range(rounds) if rounds else count()
    while proof_of_work is None:
        for _i in r:
            if proof_of_work is not None:
                break
            work = work_generator(
                unproven_header, target, proof_start, worksize, cpus
            )
            with multiprocessing.Pool(cpus) as p:
                imap = p.imap_unordered(mill_work, work)
                for (proof, c) in imap:
                    progress_next(n=c)
                    if proof is not None and proof_of_work is None:
                        proof_of_work = proof
            p.join()
            proof_start += worksize * cpus
        yield proof_of_work


def milling_generator(
    block, mp=False, rounds=None, worksize=None, progress=None
):
    rounds = rounds or 1
    worksize = worksize or 100000
    progress_next = progress.next if progress else lambda n=1: None
    milling_func = mill_block_mp if mp else mill_block
    miller = milling_func(block, rounds, worksize, progress_next)
    proof_of_work = None
    for proof_of_work in miller:
        if proof_of_work is not None:
            block.solve(proof_of_work)
        yield proof_of_work



import datetime
import uuid

from passlib.hash import argon2, pbkdf2_sha256

from cancelchain.database import db
from cancelchain.wallet import Wallet


def rollback_session():
    db.session.rollback()


block_transactions = db.Table(
    'block_transaction',
    db.Column(
        'block_id', db.Integer, db.ForeignKey('block.id'),
        primary_key=True
    ),
    db.Column(
        'transaction_id', db.Integer, db.ForeignKey('transaction.id'),
        primary_key=True
    )
)


class TransactionDAO(db.Model):
    __tablename__ = 'transaction'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    txid = db.Column(db.String(100), unique=True, nullable=False, index=True)
    version = db.Column(db.String(10), nullable=False)
    timestamp = db.Column(db.DateTime, nullable=False)
    address = db.Column(db.String(100), nullable=True)
    public_key = db.Column(db.String(500), nullable=True)
    signature = db.Column(db.String(500), nullable=True)
    blocks = db.relationship(
        'BlockDAO', secondary=block_transactions, back_populates='transactions'
    )

    def __init__(
        self, txid, version, timestamp, address=None, public_key=None,
        signature=None, inflow_daos=None, outflow_daos=None
    ):
        self.txid = txid
        self.version = version
        self.timestamp = timestamp
        self.address = address
        self.public_key = public_key
        self.signature = signature
        for inflow_dao in inflow_daos or []:
            inflow_dao.transaction = self
        for outflow_dao in outflow_daos or []:
            outflow_dao.transaction = self

    def commit(self):
        db.session.add(self)
        db.session.commit()

    @classmethod
    def get(cls, txid):
        return cls.query.filter_by(txid=txid).one_or_none()

    @classmethod
    def transactions_chain(cls, block_chain):
        block_alias = db.aliased(BlockDAO, block_chain.subquery())
        q = db.session.query(TransactionDAO)
        q = q.join(block_alias, TransactionDAO.blocks)
        return q.order_by(TransactionDAO.timestamp.desc(), TransactionDAO.id)


class OutflowDAO(db.Model):
    __tablename__ = 'outflow'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    txid = db.Column(db.String(100), nullable=False)
    idx = db.Column(db.Integer, nullable=False)
    amount = db.Column(db.BigInteger, nullable=False)
    address = db.Column(db.String(100), nullable=True)
    subject = db.Column(db.String(500), nullable=True)
    forgive = db.Column(db.String(500), nullable=True)
    support = db.Column(db.String(500), nullable=True)
    transaction_id = db.Column(
        db.Integer, db.ForeignKey('transaction.id'), nullable=False
    )
    transaction = db.relationship(
        'TransactionDAO',
        backref=db.backref('outflows', order_by='OutflowDAO.idx')
    )
    __table_args__ = (
        db.UniqueConstraint('txid', 'idx'),
        db.Index('ix_outflow_txid_idx', 'txid', 'idx'),
    )

    def __init__(
        self, txid, idx, amount,
        address=None, subject=None, forgive=None, support=None,
        transaction_dao=None
    ):
        with db.session.no_autoflush:
            self.txid = txid
            self.idx = idx
            self.amount = amount
            self.address = address
            self.subject = subject
            self.forgive = forgive
            self.support = support
            self.transaction = transaction_dao or None

    @classmethod
    def get(cls, outflow_txid, outflow_idx):
        return cls.query.filter_by(
            outflow_txid=outflow_txid, outflow_idx=outflow_idx
        ).one_or_none()

    @classmethod
    def outflows_chain(cls, transactions_chain):
        txn_alias = db.aliased(TransactionDAO, transactions_chain.subquery())
        q = db.session.query(OutflowDAO)
        q = q.join(txn_alias, OutflowDAO.transaction)
        q = q.order_by(
            txn_alias.timestamp.desc(),
            txn_alias.txid,
            OutflowDAO.idx
        )
        return q


class InflowDAO(db.Model):
    __tablename__ = 'inflow'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    txid = db.Column(db.String(100), nullable=False)
    idx = db.Column(db.Integer, nullable=False)
    outflow_txid = db.Column(db.String(100), nullable=False)
    outflow_idx = db.Column(db.Integer, nullable=False)
    outflow_id = db.Column(
        db.Integer, db.ForeignKey('outflow.id'), nullable=False
    )
    outflow = db.relationship('OutflowDAO', backref='inflows')
    transaction_id = db.Column(
        db.Integer, db.ForeignKey('transaction.id'), nullable=False
    )
    transaction = db.relationship(
        'TransactionDAO',
        backref=db.backref('inflows', order_by='InflowDAO.idx')
    )

    __table_args__ = (
        db.UniqueConstraint('txid', 'idx'),
        db.Index('ix_inflow_txid_idx', 'txid', 'idx'),
    )

    def __init__(
        self, txid, idx, outflow_txid, outflow_idx,
        outflow_dao=None, transaction_dao=None
    ):
        with db.session.no_autoflush:
            self.txid = txid
            self.idx = idx
            self.outflow_txid = outflow_txid
            self.outflow_idx = outflow_idx
            if not outflow_dao:
                outflow_dao = OutflowDAO.query.filter_by(
                    txid=outflow_txid, idx=outflow_idx
                ).one_or_none()
            self.outflow = outflow_dao
            self.transaction = transaction_dao

    @classmethod
    def inflows_chain(cls, transactions_chain):
        txn_alias = db.aliased(TransactionDAO, transactions_chain.subquery())
        q = db.session.query(InflowDAO)
        q = q.join(txn_alias, InflowDAO.transaction)
        q = q.order_by(
            txn_alias.timestamp.desc(),
            txn_alias.txid,
            InflowDAO.idx
        )
        return q


class BlockDAO(db.Model):
    __tablename__ = 'block'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    block_hash = db.Column(
        db.String(100), unique=True, nullable=False, index=True
    )
    version = db.Column(db.String(10), nullable=False)
    idx = db.Column(db.Integer, nullable=False, index=True)
    prev_hash = db.Column(db.String(100), nullable=False, index=True)
    timestamp = db.Column(db.DateTime, nullable=False)
    merkle_root = db.Column(db.String(100), nullable=False)
    proof_of_work = db.Column(db.BigInteger, nullable=False)
    target = db.Column(db.String(100), nullable=False)
    prev_id = db.Column(db.Integer, db.ForeignKey('block.id'), nullable=True)
    prev = db.relationship('BlockDAO', remote_side=[id], backref='next')
    transactions = db.relationship(
        'TransactionDAO',
        secondary=block_transactions, back_populates='blocks',
        lazy='dynamic',
        order_by=[TransactionDAO.timestamp, TransactionDAO.txid]
    )

    def __init__(
        self, block_hash, version, idx, prev_hash, timestamp, merkle_root,
        proof_of_work, target, prev_dao=None, transaction_daos=None
    ):
        self.block_hash = block_hash
        self.version = version
        self.idx = idx
        self.prev_hash = prev_hash
        self.timestamp = timestamp
        self.merkle_root = merkle_root
        self.proof_of_work = proof_of_work
        self.target = target
        self.prev = prev_dao or BlockDAO.get(prev_hash)
        for transaction_dao in transaction_daos or []:
            self.transactions.append(transaction_dao)

    @property
    def _block_chain(self):
        q = BlockDAO.query.filter(BlockDAO.id == self.id).cte(recursive=True)
        return q.union_all(BlockDAO.query.filter(BlockDAO.id == q.c.prev_id))

    @property
    def block_chain(self):
        return db.session.query(self._block_chain)

    @property
    def transactions_chain(self):
        return TransactionDAO.transactions_chain(self.block_chain)

    @property
    def outflows_chain(self):
        return OutflowDAO.outflows_chain(self.transactions_chain)

    @property
    def inflows_chain(self):
        return InflowDAO.inflows_chain(self.transactions_chain)

    def commit(self):
        db.session.add(self)
        db.session.commit()

    def get_transaction_in_chain(self, txid):
        return self.transactions_chain.filter(
            TransactionDAO.txid == txid
        ).one_or_none()

    def address_transactions(self, address):
        return self.transactions_chain.filter(
            TransactionDAO.address == address
        )

    def get_block_in_chain(self, block_hash=None, idx=None):
        block_alias = db.aliased(BlockDAO, self.block_chain.subquery())
        q = db.session.query(BlockDAO)
        q = q.join(block_alias, BlockDAO.id == block_alias.id)
        if block_hash is not None:
            q = q.filter(
                BlockDAO.block_hash == block_hash
            )
        if idx is not None:
            q = q.filter(
                BlockDAO.idx == idx
            )
        return q.one_or_none()

    def inflows_in_chain_count(self, outflow_txid, outflow_idx):
        return 1 if self.inflows_chain.filter(
            InflowDAO.outflow_txid == outflow_txid,
            InflowDAO.outflow_idx == outflow_idx
        ).first() is not None else 0

    @classmethod
    def count(cls):
        return db.session.query(db.func.count(cls.id)).one_or_none()[0]

    @classmethod
    def block_hashes(cls):
        for r in cls.query.with_entities(cls.block_hash).order_by(
            cls.timestamp.desc(), cls.block_hash
        ):
            yield r[0]

    @classmethod
    def get(cls, block_hash=None, idx=None):
        q = cls.query
        if block_hash:
            q = q.filter_by(block_hash=block_hash)
        else:
            q = q.filter_by(idx=idx)
        return q.one_or_none()


class ChainDAO(db.Model):
    __tablename__ = 'chain'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    block_hash = db.Column(
        db.String(100), unique=True, nullable=False, index=True
    )
    block_id = db.Column(
        db.Integer, db.ForeignKey('block.id'), nullable=False, index=True
    )
    block = db.relationship('BlockDAO', backref='chains')

    def __init__(self, block_hash, block_dao=None):
        self.block_hash = block_hash
        self.block = block_dao or BlockDAO.get(block_hash)

    @property
    def blocks(self):
        return self.block.block_chain

    @property
    def transactions(self):
        return self.block.transactions_chain

    @property
    def outflows(self):
        return self.block.outflows_chain

    @property
    def inflows(self):
        return self.block.inflows_chain

    def unspent_outflows(self, address, filter_pending=False):
        inflows_alias = db.aliased(InflowDAO, self.inflows.subquery())
        q = self.outflows.filter(OutflowDAO.address == address)
        q = q.join(inflows_alias, OutflowDAO.inflows, isouter=True)
        q = q.filter(inflows_alias.id.is_(None))
        if filter_pending:
            q = q.filter(~OutflowDAO.pending.any())
        return q

    def wallet_balance(self, address):
        inflows_alias = db.aliased(InflowDAO, self.inflows.subquery())
        q = self.outflows.filter(OutflowDAO.address == address)
        q = q.join(inflows_alias, OutflowDAO.inflows, isouter=True)
        q = q.filter(inflows_alias.id.is_(None))
        outflows_alias = db.aliased(OutflowDAO, q.subquery())
        q2 = db.session.query(
            db.func.sum(OutflowDAO.amount)
        ).join(outflows_alias, OutflowDAO.id == outflows_alias.id)
        amount = q2.one_or_none()
        return (amount[0] or 0) if amount is not None else 0

    def unforgiven_outflows(self, subject, address=None, filter_pending=False):
        inflows_alias = db.aliased(InflowDAO, self.inflows.subquery())
        q = self.outflows.filter(OutflowDAO.subject == subject)
        q = q.join(inflows_alias, OutflowDAO.inflows, isouter=True)
        q = q.filter(inflows_alias.id.is_(None))
        if address is not None:
            txn_alias = db.aliased(
                TransactionDAO, self.transactions.subquery()
            )
            q = q.join(txn_alias, OutflowDAO.transaction)
            q = q.filter(txn_alias.address == address)
        if filter_pending:
            q = q.filter(~OutflowDAO.pending.any())
        return q

    def subject_balance(self, subject):
        inflows_alias = db.aliased(InflowDAO, self.inflows.subquery())
        q = self.outflows.filter(OutflowDAO.subject == subject)
        q = q.join(inflows_alias, OutflowDAO.inflows, isouter=True)
        q = q.filter(inflows_alias.id.is_(None))
        outflows_alias = db.aliased(OutflowDAO, q.subquery())
        q2 = db.session.query(
            db.func.sum(OutflowDAO.amount)
        ).join(outflows_alias, OutflowDAO.id == outflows_alias.id)
        amount = q2.one_or_none()
        return (amount[0] or 0) if amount is not None else 0

    def subject_support(self, subject):
        q = self.outflows.filter(OutflowDAO.support == subject)
        outflows_alias = db.aliased(OutflowDAO, q.subquery())
        q2 = db.session.query(
            db.func.sum(OutflowDAO.amount)
        ).join(outflows_alias, OutflowDAO.id == outflows_alias.id)
        amount = q2.one_or_none()
        return (amount[0] or 0) if amount is not None else 0

    def wallet_leaderboard(self, earliest=None, latest=None, limit=None):
        inflows_alias = db.aliased(InflowDAO, self.inflows.subquery())
        txn_alias = db.aliased(TransactionDAO, self.transactions.subquery())
        q = db.session.query(
            OutflowDAO.address, db.func.sum(OutflowDAO.amount).label('ct')
        )
        q = q.filter(OutflowDAO.address.is_not(None))
        q = q.join(txn_alias, OutflowDAO.transaction)
        q = q.join(inflows_alias, OutflowDAO.inflows, isouter=True)
        q = q.filter(inflows_alias.id.is_(None))
        if earliest is not None:
            q = q.filter(txn_alias.timestamp >= earliest)
        if latest is not None:
            q = q.filter(txn_alias.timestamp < latest)
        q = q.group_by(OutflowDAO.address)
        q = q.order_by(db.desc('ct'), OutflowDAO.address)
        if limit is not None:
            q = q.limit(limit)
            return db.session.query(db.aliased(q.subquery()))
        return q

    def set_block_hash(self, block_hash):
        self.block = BlockDAO.get(block_hash)
        self.block_hash = block_hash

    def get_block(self, block_hash=None, idx=None):
        return self.block.get_block_in_chain(block_hash=block_hash, idx=idx)

    def next_block(self, block):
        for next_block in block.next:
            if self.get_block(block_hash=next_block.block_hash) is not None:
                return next_block
        return None

    def get_transaction(self, txid):
        return self.block.get_transaction_in_chain(txid)

    def address_transactions(self, address):
        return self.block.address_transactions(address)

    def commit(self):
        db.session.add(self)
        db.session.commit()

    @classmethod
    def count(cls):
        return db.session.query(db.func.count(cls.id)).one_or_none()[0]

    @classmethod
    def get(cls, block_hash=None, id=None):
        q = cls.query
        if block_hash:
            q = q.filter_by(block_hash=block_hash)
        else:
            q = q.filter_by(id=id)
        return q.one_or_none()

    @classmethod
    def ids(cls):
        for r in cls.query.with_entities(cls.id).order_by(cls.id):
            yield r[0]

    @classmethod
    def chains(cls):
        return cls.query.join(cls.block).order_by(
            BlockDAO.idx.desc(), BlockDAO.timestamp, BlockDAO.block_hash
        )

    @classmethod
    def longest(cls):
        return cls.chains().first()


class PendingTxnDAO(db.Model):
    __tablename__ = 'pending_txn'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    txid = db.Column(db.String(100), unique=True, nullable=False, index=True)
    timestamp = db.Column(db.DateTime, nullable=False)
    json_data = db.Column(db.Text(), nullable=False)
    received = db.Column(db.DateTime, default=datetime.datetime.utcnow)

    def add(self):
        db.session.add(self)

    def commit(self):
        self.add()
        db.session.commit()

    def delete(self):
        db.session.delete(self)
        db.session.commit()

    @classmethod
    def count(cls):
        return db.session.query(db.func.count(cls.id)).one_or_none()[0]

    @classmethod
    def json_datas(cls, earliest=None, expired=None):
        q = cls.query.with_entities(cls.json_data)
        if earliest is not None:
            q = q.filter(cls.received >= earliest)
        if expired is not None:
            q = q.filter(cls.timestamp >= expired)
        q = q.order_by(cls.timestamp, cls.txid)
        for r in q:
            yield r[0]

    @classmethod
    def get(cls, txid):
        return cls.query.filter_by(txid=txid).one_or_none()


class PendingIOflowDAO(db.Model):
    __tablename__ = 'pending_ioflow'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    txid = db.Column(db.String(100), nullable=False)
    outflow_txid = db.Column(db.String(100), nullable=False)
    outflow_idx = db.Column(db.Integer, nullable=False)
    pending_txn_id = db.Column(
        db.Integer, db.ForeignKey('pending_txn.id'), nullable=False
    )
    pending_txn = db.relationship(
        'PendingTxnDAO',
        backref=db.backref('ioflows', cascade='delete, delete-orphan')
    )
    outflow_id = db.Column(
        db.Integer, db.ForeignKey('outflow.id'), nullable=False
    )
    outflow = db.relationship('OutflowDAO', backref='pending')

    def add(self):
        db.session.add(self)

    def commit(self):
        self.add()
        db.session.commit()


class ChainFill(db.Model):
    __tablename__ = 'chain_fill'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)

    def add(self):
        db.session.add(self)

    def commit(self):
        self.add()
        db.session.commit()

    def delete(self):
        db.session.delete(self)
        db.session.commit()


class ChainFillBlock(db.Model):
    __tablename__ = 'chain_fill_block'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    block_hash = db.Column(db.String(100), nullable=False)
    idx = db.Column(db.Integer, nullable=False)
    block_json = db.Column(db.Text, nullable=True)
    chain_fill_id = db.Column(
        db.Integer, db.ForeignKey('chain_fill.id'), nullable=False
    )
    chain_fill = db.relationship(
        'ChainFill',
        backref=db.backref(
            'blocks',
            order_by='ChainFillBlock.idx',
            cascade='delete, delete-orphan'
        )
    )

    def add(self):
        db.session.add(self)

    def commit(self):
        self.add()
        db.session.commit()


class ApiToken(db.Model):
    __tablename__ = 'api_token'

    id = db.Column(db.Integer, autoincrement=True, primary_key=True)
    address = db.Column(
        db.String(100), unique=True, nullable=False, index=True
    )
    public_key = db.Column(db.String(500), nullable=False)
    hashed = db.Column(db.String(100), unique=True, nullable=True)
    cipher = db.Column(db.String(500), unique=True, nullable=True)
    timestamp = db.Column(
        db.DateTime,
        default=datetime.datetime.utcnow,
        onupdate=datetime.datetime.utcnow
    )

    @property
    def expired(self):
        now_dt = datetime.datetime.now(datetime.timezone.utc)
        now_dt = now_dt.replace(tzinfo=None)
        return self.timestamp < (now_dt - datetime.timedelta(seconds=60))

    def add(self):
        db.session.add(self)

    def commit(self):
        self.add()
        db.session.commit()

    def refreshed_cipher(self):
        if self.expired or not (self.cipher and self.hashed):
            secret = str(uuid.uuid4())
            try:
                self.hashed = argon2.hash(secret)
            except Exception:
                self.hashed = pbkdf2_sha256.hash(secret)
            wallet = Wallet(b64ks=self.public_key)
            self.cipher = wallet.encrypt(secret.encode())
            self.commit()
        return self.cipher

    def reset(self):
        self.cipher = None
        self.hashed = None
        self.commit()

    def verify(self, secret):
        hm = argon2 if argon2.identify(self.hashed) else pbkdf2_sha256
        return hm.verify(secret, self.hashed) and not self.expired

    @classmethod
    def get(cls, address):
        return cls.query.filter_by(address=address).one_or_none()

    @classmethod
    def create(cls, wallet):
        api_token = cls(
            address=wallet.address, public_key=wallet.public_key_b64
        )
        api_token.commit()
        return api_token
